/*
 * newlogin ist ein simples programm, welches /etc/utmp beobachtet und
 * an- und abgaenge von usern meldet...
 *
 * (c) Copyright 1990 by A.Zinser   fifi@veeble.uucp
 */
 
#include	<stdio.h>
#include	<sys/types.h>
#include	<utmp.h>
#include	<time.h>
#include	<signal.h>
#include	<pwd.h>

#define		INTERACT 

#define		NEWLOGIN	".newlogin"
#define		NEWLOGINPATH	"/usr/local/lib"
#define 	LOGDATEI	"/usr/adm/logins"
#define		UUCICO		"uucico"
#define 	CYCLE		5

extern	char	*calloc();
extern	int	getopt();
extern	char	*optarg;
extern	int	optind;
extern	char	*strchr();
extern	char	*strtok();

#ifndef	UTMP_FILE
	#define	UTMP_FILE	"/etc/utmp"
#endif
#define	MAXUSER	90

#define	strsave(x)	((char *)strcpy(ckalloc(1, (strlen(x) + 1)), (x)))

struct ttys {
	struct ttys *next;
	char	*dev;
	char	*descr;
} *root_ttys, *read_newlogin();

static int		modus = 0;
#define	QUIET		16
#define	USERS		32	/* output like bsd users */
#define	NPERM		8
#define	OPTM		4
#define	OPTI		2
#define	OPTD		1
#define	DEBUG		64
#define	WISHTOSEE	!(modus & OPTD)
#define	DONTWONNASEE	(modus & OPTD)

char *
ckalloc(n,s)
unsigned long n,s;
{
	char *cp;
	if (!(cp = calloc(n,s))) {
		fprintf(stderr,"No more memory... bye..\n");
		exit(-1);
	}
	return(cp);
}

main (argc,argv)
int	argc;		/* 1 = 0 args */
char	*argv[];
{
	FILE		*fputmp;
	struct	utmp	utar[MAXUSER];
	int		i,j,anzuser = 0, usercnt = 0;
	int		opterg;
	struct passwd	*passwd, *getpwnam();
	char		*cp;
	int		cycle = CYCLE;
#ifdef INTERACT
	char		iline[80];
	char		*user[2*MAXUSER];
#else
	char		**user;
#endif

	if (!strcmp(argv[0],"users")) modus |= USERS;

	while((opterg = getopt(argc,argv,"xDuqpmit:hd")) != EOF) {
		switch(opterg) {
		default:
		case '?':
		case 'h':
			printf("Usage: %s [-t cycle] [-d] device-/userlist\n",argv[0]);
			printf("	-t cycle   Set watchcycle to 'cycle' seconds\n");
			printf("	-u         output like bsd4.3 users\n");
			printf("	-q         don't display users who are currently logged in\n");
			printf("	-d         Ignore all following Users/ttys\n");
			printf("	-p         stay permanent in background and watch /etc/utmp\n");
			printf("	-m         don't display polling sites\n");
			printf("	list       Only watch/ignore these users/ttys\n");
			exit(0);
		case 'D':
			modus |= DEBUG;
			break;
		case 'u':
			modus |= USERS;
			break;
		case 'q':
			modus |= QUIET;
			break;
		case 'p':
			modus |= NPERM;
			break;
		case 'd':
			modus |= OPTD;
			break;
		case 't':
			if (!(cycle=atoi(optarg))) cycle = CYCLE;
			break;
		case 'm':
			modus |= OPTM;
			break;
#ifdef INTERACT
		case 'i':
			modus |= OPTI;
			break;
		}
#endif
	}

	if (optind<argc) {
#ifndef INTERACT
		user  = (char **)ckalloc((argc-optind), sizeof(char *));
#endif
		for(; optind<argc; optind++) {
			if ((passwd = getpwnam(argv[optind])) || (!strncmp(argv[optind],"tty",3)))
				user[usercnt++] = strsave(argv[optind]);
			else
				fprintf(stderr,"Unbekannter Benutzer %s\n",argv[optind]);
		}
	}
#ifdef INTERACT
	if (modus & 2) {
		printf("Args?: "); fflush(stdout);
		if (gets(iline)) {
			cp = strtok(iline," ");
			if(!strcmp(cp,"-d")) {
				modus |= 1;
				cp = strtok(NULL," ");
			}
			while(cp) {
				if ((passwd = getpwnam(cp)) || (!strncmp(cp,"tty",3)))
					user[usercnt++] = strsave(cp);
				else
					fprintf(stderr,"Unbekannter Benutzer %s\n",cp);
				cp = strtok(NULL," ");
			}
		}
	}
#endif

	root_ttys = read_newlogin();

	if ((fputmp = fopen(UTMP_FILE, "r")) == NULL) {
		fprintf(stderr, "Cannot open %s\n", UTMP_FILE);
		exit(1);
	}

	/* erstmaliges lesen von utmp */

	for(i=1; i<MAXUSER && fread((char *)&utar[i], sizeof(struct utmp), 1, fputmp); i++);
	anzuser = i;

	if (!(modus & QUIET)) {
		for (i=1; i<anzuser; i++) {
			if (modus & DEBUG) {
				printf("%d %d/%d %-8.8s %-8.8s\n",utar[i].ut_type, utar[i].ut_exit.e_termination, utar[i].ut_exit.e_exit, utar[i].ut_user, utar[i].ut_line);
			}
			if (utar[i].ut_type == USER_PROCESS) {
				if (usercnt) for (j=usercnt; j && strcmp(utar[i].ut_name, user[j-1]) && strcmp(utar[i].ut_line, user[j-1]); j--);
				if (!usercnt || (j && WISHTOSEE) || (!j && DONTWONNASEE))  {
					if (modus & USERS) printf("%-.8s \n",utar[i].ut_name);
					else pruser(&utar[i],'*');
				}
				/* pruser(&utar[i],'*'); */
			}
		}
		if (modus & USERS) printf("\n");
	}
	if (!strncmp(argv[0],"wx",2) || !(modus & NPERM))
		return(0);

	if (fork()) {
		return(0);
	}

	signal(SIGINT, SIG_IGN);

	while(1) {
		sleep(cycle);
		rewind(fputmp);
		for(i=1; i<anzuser && fread((char *)&utar[0], sizeof(struct utmp), 1, fputmp); i++) {
			if (utar[0].ut_time != utar[i].ut_time) {
				if (modus & DEBUG) {
					printf("%d %d/%d %-8.8s %-8.8s\n",utar[0].ut_type, utar[0].ut_exit.e_termination, utar[0].ut_exit.e_exit, utar[0].ut_user, utar[0].ut_line);
				}
				if (utar[0].ut_type == USER_PROCESS) {
					if (usercnt) for (j=usercnt; j && strcmp(utar[0].ut_name, user[j-1]) && strcmp(utar[i].ut_line, user[j-1]); j--);
					if (!usercnt || (j && WISHTOSEE) || (!j && DONTWONNASEE))  pruser(&utar[0],'+');
				}
				else {
					if (utar[i].ut_type == USER_PROCESS) {
						if (usercnt) for (j=usercnt; j && strcmp(utar[i].ut_name, user[j-1]) && strcmp(utar[i].ut_line, user[j-1]); j--);
						if (!usercnt || (j && WISHTOSEE) || (!j && DONTWONNASEE))  pruser(&utar[i],'-');
					}
				}
			}
			memcpy(&utar[i], &utar[0], sizeof(struct utmp));
		}
	}	
	exit (fclose(fputmp));
}

pruser(utmpp,mode)
struct	utmp	*utmpp;
char		mode;
{
	FILE		*fputlo;
	long   clock;
	struct tm *timent;
	struct tm *timent2;
	struct passwd *passwd;
	static char	user[255];
	static char	user1[255];
	static char	dev[255];
	char	*shell;
	char	*cp;
	struct ttys	*ttyp = root_ttys;
	int		wrprot;
	
	if (modus & NPERM)   {
	if ((fputlo = fopen(LOGDATEI, "a")) == NULL) {
		fprintf(stderr, "Cannot open %s\n", LOGDATEI);
		exit(1); }
	}

	sprintf(dev,"/dev/%s",utmpp->ut_line);
	wrprot = access(dev,02);

	while(ttyp && strcmp(ttyp->dev,utmpp->ut_line)) ttyp = ttyp->next;
	if(ttyp)
		strcpy(dev,ttyp->descr);
	else
		sprintf(dev,"/dev/%s",utmpp->ut_line);

	strncpy(user,utmpp->ut_name,8);
	user[8] = '\0';
	passwd = getpwnam(user);
	if (passwd) {
		cp = strcpy(user1,passwd->pw_gecos);
		while(*cp && (*cp != ',') && (*cp != ':')) cp++; *cp = '\0';
		sprintf(user,"%s (%s)",user1,passwd->pw_name);
		shell = passwd->pw_shell;
	}
	else {
		shell = "/bin/sh";
	}
	cp = strrchr(shell,'/');
	if ((modus & OPTM) && cp && !strcmp((cp+1),UUCICO))
		return;

	timent = localtime(&utmpp->ut_time);
	clock = time(0L);
	timent2 = localtime(&clock);
	if (modus & NPERM)   {
	if (mode ==  '-')
	fprintf(fputlo,"%c %02d.%02d %02d:%02d:%02d %s%s [%s]\n\r",
		mode,
		timent2->tm_mday,
		(timent2->tm_mon + 1),
		/* timent2->tm_year, */
		timent2->tm_hour,
		timent2->tm_min,
		timent2->tm_sec, 
		(wrprot ? "(*) " : ""),
		user,
		dev
		);
	else
	fprintf(fputlo,"%c %02d.%02d %02d:%02d:%02d %s%s [%s]\n\r",
		mode,
		timent->tm_mday,
		(timent->tm_mon + 1),
		/* timent->tm_year, */
		timent->tm_hour,
		timent->tm_min,
		timent->tm_sec, 
		(wrprot ? "(*) " : ""),
		user,
		dev
		);
	}
	else
	{
	printf("%c %02d.%02d %02d:%02d:%02d %s%s [%s]\n\r",
		mode,
		timent->tm_mday,
		(timent->tm_mon + 1),
		/* timent->tm_year, */
		timent->tm_hour,
		timent->tm_min,
		timent->tm_sec, 
		(wrprot ? "(*) " : ""),
		user,
		dev
		);
	}

	if (modus & NPERM)   {
	fclose(fputlo);
	}
}

FILE *
open_newlogin()
{
	char	fname[255];
	FILE	*fp;

	if ((fp = fopen(NEWLOGIN,"r")) != NULL)
		return(fp);
	strcpy(fname,getenv("HOME"));
	strcat(fname,"/");
	strcat(fname,NEWLOGIN);
	if ((fp = fopen(fname,"r")) != NULL)
		return(fp);
	strcpy(fname,NEWLOGINPATH);
	strcat(fname,"/");
	strcat(fname,NEWLOGIN);
	if ((fp = fopen(fname,"r")) != NULL)
		return(fp);
	return(NULL);
}

struct ttys *
read_newlogin()
{
	char	*cp;
	FILE	*fp = open_newlogin();
	char	buf[255];
	struct	ttys	*rttyp, *ttyp = NULL;

	if (!fp) {
		fprintf(stderr,"Datei %s nicht gefunden\n",NEWLOGIN);
		return NULL;
	}

	while(fgets(buf, sizeof(buf)-1, fp) != NULL) {
		if ((cp = strchr(buf,'#')) && *(cp+1)) {
			*(buf + strlen(buf) - 1) = '\0';
			if (!ttyp) {
				rttyp = ttyp = (struct ttys *)ckalloc(1, sizeof(struct ttys));
				ttyp->next = NULL;
			} else {
				ttyp->next = (struct ttys *)ckalloc(1, sizeof(struct ttys));
				ttyp = ttyp->next;
				ttyp->next = NULL;
			}
			ttyp->descr = strsave(cp+1);
			*cp = 0;
			ttyp->dev = strsave(buf);
		} else {
			fprintf(stderr,"Fehlerhafte Zeile in %s: %s\n",NEWLOGIN,buf);
		}
	}

	fclose(fp);
	return(rttyp);
}
